import math
import IsoSpecPy
import numpy as np
from scipy.stats import norm, uniform, gamma
import random
import heapq
import re
from collections import Counter
import numpy.random as rd
from scipy.signal import argrelmax
from .peptides import get_protein_formula
from warnings import warn
from copy import deepcopy

class Spectrum:
    def __init__(self, formula='', threshold=0.001, total_prob=None,
                 charge=1, adduct=None, confs=None, label=None, **other):
        """Initialize a Spectrum class.

        Initialization can be done either by simulating a spectrum of an ion
        with a given formula and charge, or setting a peak list.

        The initialized spectrum is not normalised. In order to do this use
        normalize method.

        Parameters
        ----------

        formula: str
            The chemical formula of the molecule. If empty, then `confs`
            cannot be None. If the formula is a valid chemical formula then
            spectrum peaks (`confs`) are simulated.
        threshold: float
            Lower threshold on the intensity of simulated peaks. Used when
            `formula` is not an empty string, ignored when `total_prob` is not
            None.
        total_prob: float
            Lower bound on the total probability of simulated peaks, i.e.
            fraction of all potential signal, which will be simulated. Used
            when `formula` is not an empty string. When not None, then
            `threshold` value is ignored.
        charge: int
            A charge of the ion.
        adduct: str
            The ionizing element. When not None, then formula is updated
            with `charge` number of adduct atoms.
        confs: list
            A list of tuples of mz and intensity. Confs contains peaks of an
            initialized spectrum. If not None, then `formula` needs be an empty
            string.
        label: str
            An optional spectrum label.
        """
        ### TODO2: seprarate subclasses for centroid & profile spectra
        self.formula = formula
        self.empty = False

        if label is None:
            self.label = formula
        else:
            self.label = label

        self.charge = charge

        if formula != '' and confs is not None:
            raise ValueError(
                "Formula and confs cannot be set at the same time!")
        elif confs is not None:
            self.set_confs(confs)
        elif formula != '':
            self.set_confs(
                self.confs_from_formula(
                    formula, threshold, total_prob, charge, adduct))
        else:
            self.empty = True
            self.confs = []

    @staticmethod
    def confs_from_formula(formula, threshold=0.001, total_prob=None,
                           charge=1, adduct=None):
        """Simulate and return spectrum peaks for given formula.

        Parameters as in __init__ method. `formula` must be a nonempty string.
        """
        parsed = re.findall('([A-Z][a-z]*)([0-9]*)', formula)
        formula = Counter()
        for e, n in parsed:
            n = int(n) if n else 1
            formula[e] += n
        if adduct:
            formula[adduct] += charge
        assert all(v >= 0 for v in formula.values())
        formula = ''.join(x+str(formula[x]) for x in formula if formula[x])
        if total_prob is not None:
            isospec = IsoSpecPy.IsoTotalProb(formula=formula,
                                             prob_to_cover=total_prob,
                                             get_minimal_pset=True,
                                             get_confs=False)
        else:
            isospec = IsoSpecPy.IsoThreshold(formula=formula,
                                             threshold=threshold,
                                             absolute=False,
                                             get_confs=False)
        confs = [(x[0]/abs(charge), x[1]) for x in
                 zip(isospec.masses, isospec.probs)]
        return confs

    @staticmethod
    def new_from_fasta(fasta, threshold=0.001, total_prob=None, intensity=1.0,
                       empty=False, charge=1, label=None):
        return Spectrum(get_protein_formula(fasta), threshold=threshold,
                        total_prob=total_prob, intensity=intensity,
                        empty=empty, charge=charge, label=label)

    @staticmethod
    def new_from_csv(filename, delimiter=","):
        spectrum = Spectrum(label=filename)

        with open(filename, "r") as infile:
            header = next(infile)
            for line in infile:
                if line[0] == '#':
                    continue
                line = line.strip()
                line = line.split(delimiter)
                spectrum.confs.append(tuple(map(float, line)))
        spectrum.sort_confs()
        spectrum.merge_confs()
        return spectrum

    @staticmethod
    def new_random(domain=(0.0, 1.0), peaks=10):
        ret = Spectrum()
        confs = []
        for _ in range(peaks):
            confs.append((random.uniform(*domain), random.uniform(0.0, 1.0)))
        ret.set_confs(confs)
        return ret

    def average_mass(self):
        """
        Returns the average mass.
        """
        norm = float(sum(x[1] for x in self.confs))
        return sum(x[0]*x[1]/norm for x in self.confs)

    def copy(self):
        """
        Return a (deep) copy of self
        """
        return deepcopy(self)

    # def copy(self):
    #     isospec = self.isospec
    #     self.isospec = None
    #     ret = deepcopy(self)
    #     ret.isospec = isospec
    #     self.isospec = isospec
    #     return ret

    def get_modal_peak(self):
        """
        Returns the peak with the highest intensity.
        """
        return max(self.confs, key=lambda x: x[1])

    def sort_confs(self):
        """
        Sorts configurations by their mass.
        """
        self.confs.sort(key = lambda x: x[0])

    def merge_confs(self):
        """
        Merges configurations with an identical mass, summing their intensities.
        """
        if not self.empty:
            cmass = self.confs[0][0]
            cprob = 0.0
            ret = []
            for mass, prob in self.confs + [(-1, 0)]:
                if mass != cmass:
                    ret.append((cmass, cprob))
                    cmass = mass
                    cprob = 0.0
                cprob += prob
            ### TODO3: for profile spectra, set a margin of max. 5 zero intensities
            ### around any observed intensity to preserve peak shape
            ### For centroid spectra, remove all zero intensities.
            #self.confs = [x for x in ret if x[1] > 1e-12]
            self.confs = ret

    def set_confs(self, confs):
        self.confs = confs
        if len(self.confs) > 0:
            self.sort_confs()
            self.merge_confs()
        else:
            self.empty = True

    def __add__(self, other):
        res = Spectrum()
        res.confs = self.confs + other.confs
        res.sort_confs()
        res.merge_confs()
        res.label = self.label + ' + ' + other.label
        return res

    def __mul__(self, number):
        res = Spectrum()
        res.set_confs([(x[0], number*x[1]) for x in self.confs])
        res.label = self.label
        return res

    def __rmul__(self, number):
        # Here * is commutative
        return self * number

    def __len__(self):
        return len(self.confs)

    @staticmethod
    def ScalarProduct(spectra, weights):
        ret = Spectrum()
        Q = [(spectra[i].confs[0], i, 0) for i in range(len(spectra))]
        heapq.heapify(Q)
        while Q != []:
            conf, spectre_no, conf_idx = heapq.heappop(Q)
            ret.confs.append((conf[0], conf[1] * weights[spectre_no]))
            conf_idx += 1
            if conf_idx < len(spectra[spectre_no]):
                heapq.heappush(Q, (spectra[spectre_no].confs[conf_idx], spectre_no, conf_idx))
        ret.merge_confs()
        return ret

    def normalize(self, target_value = 1.0):
        """
        Normalize the intensity values so that they sum up to the target value.
        """
        x = target_value/math.fsum(v[1] for v in self.confs)
        self.confs = [(v[0], v[1]*x) for v in self.confs]

    def WSDistanceMoves(self, other):
        """
        Return the optimal transport plan between self and other.
        """
        try:
            ii = 0
            leftoverprob = other.confs[0][1]
            for mass, prob in self.confs:
                while leftoverprob <= prob:
                    yield (other.confs[ii][0], mass, leftoverprob)
                    prob -= leftoverprob
                    ii += 1
                    leftoverprob = other.confs[ii][1]
                yield (other.confs[ii][0], mass, prob)
                leftoverprob -= prob
        except IndexError:
            return

    def WSDistance(self, other):
        if not np.isclose(sum(x[1] for x in self.confs), 1.):
            raise ValueError('Self is not normalized.')
        if not np.isclose(sum(x[1] for x in other.confs), 1.):
            raise ValueError('Other is not normalized.')
        return math.fsum(abs(x[0]-x[1])*x[2] for x in self.WSDistanceMoves(other))

    def explained_intensity(self,other):
        """
        Returns the amount of mutual intensity between self and other,
        defined as sum of minima of intensities, mass-wise.
        """
        e = 0
        for i in range(len(self.confs)):
            e += min(self.confs[i][1],other.confs[i][1])
        return e

    def bin_to_nominal(self, nb_of_digits=0):
        """
        Rounds mass values to a given number of decimal digits.
        Works in situ, returns None.
        The masses are multiplied by the charge prior to rounding,
        and divided by the charge again after rounding.
        The default nb_of_digits is zero, meaning that the m/z values
        will correspond to nominal mass of peaks.
        """
        xcoord, ycoord = zip(*self.confs)
        xcoord = map(lambda x: x*self.charge, xcoord)
        xcoord = (xcoord[0] + round(x-xcoord[0], nb_of_digits) for x in xcoord)
        xcoord = map(lambda x: x/self.charge, xcoord)
        self.confs = list(zip(xcoord, ycoord))
        self.sort_confs()
        self.merge_confs()

    def coarse_bin(self, nb_of_digits):
        """
        Rounds the m/z to a given number of decimal digits
        """
        self.confs = [(round(x[0], nb_of_digits), x[1]) for x in self.confs]
        self.merge_confs()

    def add_chemical_noise(self, nb_of_noise_peaks, noise_fraction, span=1.2):
        """
        Add additional peaks that simulate chemical noise.

        The method adds additional peaks with uniform distribution in the m/z
        domain and gamma distribution in the intensity domain. The spectrum
        does NOT need to be normalized. Accordingly, the method does not
        normalize the intensity afterwards! Works in-situ (on self).

        Parameters
        ----------
        nb_of_noise_peaks : int
            The number of added peaks.
        noise_fraction : float
            The amount of noise signal in the spectrum, >= 0 and <= 1.
        span: float or 2-tuple of floats
           If float, then `span` specifies a factor by which the m/z range is
           increased. If 2-tuple, then `span` specifies m/z range, which is
           noised.

        Returns
        -------
        None
        """
        if isinstance(span, (float, int)):
            span_increase = span
            prev_span = (min(x[0] for x in self.confs),
                         max(x[0] for x in self.confs))
            span_move = 0.5 * (span_increase - 1) * (prev_span[1] - prev_span[0])
            span = (max(prev_span[0] - span_move, 0),
                    prev_span[1] + span_move)
        noisex = uniform.rvs(loc=span[0], scale=span[1]-span[0],
                             size=nb_of_noise_peaks)
        noisey = gamma.rvs(a=2, scale=2, size=nb_of_noise_peaks)
        noisey /= sum(noisey)
        signal = sum(x[1] for x in self.confs)
        noisey *=  signal * noise_fraction / (1 - noise_fraction)
        noise = [(x, y) for x,y in zip(noisex, noisey)]
        self.confs.extend(noise)
        self.sort_confs()
        self.merge_confs()

    def add_gaussian_noise(self, sd):
        """
        Adds gaussian noise to each peak, simulating
        electronic noise.
        """
        noised = rd.normal([y for x,y in self.confs], sd)
        # noised = noised - min(noised)
        self.confs = [(x[0], y) for x, y in zip(self.confs, noised) if y > 0]

    def distort_intensity(self, N, gain, sd):
        """
        Distorts the intensity measurement in a mutiplicative noise model - i.e.
        assumes that each ion yields a random amount of signal.
        Assumes the molecule is composed of one element, so it's
        an approximation for normal molecules.
        The resulting spectrum is not normalized.
        Works in situ (modifies self).
        N: int
            number of ions
        gain: float
            mean amount of signal of one ion
        sd: float
            standard deviation of one ion's signal

        Return: np.array
            The applied deviations.
        """
        p = np.array([x[1] for x in self.confs])
        assert np.isclose(sum(p), 1), 'Spectrum needs to be normalized prior to distortion'
        X = [(x[0], N*gain*x[1]) for x in self.confs]  # average signal
        peakSD = np.sqrt(N*sd**2*p + N*gain**2*p*(1-p))
        U = rd.normal(0, 1, len(X))
        U *= peakSD
        X = [(x[0], max(x[1] + u, 0.)) for x, u in zip(X, U)]
        self.confs = X
        return U

    def distort_mz(self, mean, sd):
        """
        Distorts the m/z measurement by a normally distributed
        random variable with given mean and standard deviation.
        Use non-zero mean to approximate calibration error.
        Returns the applied shift.
        """
        N = rd.normal(mean, sd, len(self.confs))
        self.confs = [(x[0] + u, x[1]) for x, u in zip(self.confs, N)]
        self.sort_confs()
        self.merge_confs()
        return N

    @staticmethod
    def sample_multinomial(reference, N, gain, sd):
        """
        Samples a spectrum of N molecules based on peak probabilities
        from the reference spectrum. Simulates both isotope composition
        and amplifier randomness.
        The returned spectrum is not normalized.
        N: int
            number of ions in the spectrum
        gain: float
            The gain of the amplifier, i.e. average signal from one ion
        sd: float
            Standard deviation of one ion's signal
        """
        p = [x[1] for x in reference.confs]
        assert np.isclose(sum(p), 1), 'Spectrum needs to be normalized prior to sampling'
        U = rd.multinomial(N, p)
        U = rd.normal(U*gain, np.sqrt(U*sd**2))
        retSp = Spectrum('', empty=True, label='Sampled ' + reference.label)
        retSp.set_confs([(x[0], max(u, 0.)) for x, u in zip(reference.confs, U)])
        return retSp

    def find_peaks(self):
        """
        Returns a list of local maxima.
        Each maximum is reported as a tuple of m/z and intensity.
        The last and final configuration is never reported as a maximum.
        Note that this function should only be applied to profile spectra - the result
        does not make sense for centroided spectrum.
        Applying a gaussian or Savitzky-Golay filter prior to peak picking
        is advised in order to avoid detection of noise.
        """
        diffs = [n[1]-p[1] for n,p in zip(self.confs[1:], self.confs[:-1])]
        is_max = [nd <0 and pd > 0 for nd, pd in zip(diffs[1:], diffs[:-1])]
        peaks = [x for x, p in zip(self.confs[1:-1], is_max) if p]
        return peaks

    def trim_negative_intensities(self):
        """
        Detects negative intensity measurements and sets them to 0.
        """
        self.confs = [(mz, intsy if intsy >= 0 else 0.) for mz, intsy in self.confs]

    def centroid(self, max_width, peak_height_fraction=0.5):
        """Return confs of a centroided spectrum.

        The function identifies local maxima of intensity and integrates peaks in the regions
        delimited by peak_height_fraction of the apex intensity.
        By default, for each peak the function will integrate the region delimited by the full width at half maximum.
        If the detected region is wider than max_width, the peak is considered as noise and discarded.
        Small values of max_width tend to miss peaks, while large ones increase computational complexity
        and may lead to false positives.

        Note that this function should only be applied to profile spectra - the result
        does not make sense for centroided spectrum.
        Applying a gaussian or Savitzky-Golay filter prior to peak picking
        is advised in order to avoid detection of noise.

        Returns
        -----------------
            A tuple of two peak lists that can be used to construct a new Spectrum object.
            The first list contains configurations of centroids (i.e. centers of mass and areas of peaks).
            The second list contains configurations of peak apices corresponding to the centroids
            (i.e. locations and heights of the local maxima of intensity.)
        """
        ### TODO: change max_width to be in ppm?
        # Validate the input:
        if any(intsy < 0 for mz, intsy in self.confs):
            warn("""
                 The spectrum contains negative intensities!
                 It is advised to use Spectrum.trim_negative_intensities() before any processing
                 (unless you know what you're doing).
                 """)

        # Transpose the confs list to get an array of masses and an array of intensities:
        mz, intsy = np.array(self.confs).T

        # Find the local maxima of intensity:
        peak_indices = argrelmax(intsy)[0]

        peak_mz = []
        peak_intensity = []
        centroid_mz = []
        centroid_intensity = []
        max_dist = max_width/2.
        n = len(mz)
        for p in peak_indices:
            current_mz = mz[p]
            current_intsy = intsy[p]
            # Compute peak centroids:
            target_intsy = peak_height_fraction*current_intsy
            right_shift = 1
            left_shift = 1
            # Get the mz points bounding the peak fragment to integrate.
            # First, go to the right from the detected apex until one of the four conditions are met:
            # 1. we exceed the mz range of the spectrum
            # 2. we exceed the maximum distance from the apex given by max_dist
            # 3. the intensity exceeds the apex intensity (meaning that we've reached another peak)
            # 4. we go below the threshold intensity (the desired stopping condition)
            # Note: in step 3, an alternative is to check if the intensity simply starts to increase w.r.t. the previous inspected point.
            # Such an approach may give less false positive peaks, but is very sensitive to electronic noise and to overlapping peaks.
            # When we check if the intensity has not exceeded the apex intensity, and we encounter a cluster of overlapping peaks,
            # then we will effectively consider the highest one as the true apex of the cluster and integrate the whole cluster only once.
            while p + right_shift < n-1 and mz[p+right_shift] - mz[p] < max_dist and intsy[p+right_shift] <= current_intsy and intsy[p+right_shift] > target_intsy:
                right_shift += 1
            # Get the mz values of points around left mz value of the peak boundary (which will be interpolated):
            rx1, rx2 = mz[p+right_shift-1], mz[p+right_shift]
            ry1, ry2 = intsy[p+right_shift-1], intsy[p+right_shift]
            if not ry1 >= target_intsy >= ry2:
                # warn('Failed to find the right boundary of the peak at %f (probably found an overlapping peak)' % current_mz)
                continue
            # Find the left boundary of the peak:
            while p - left_shift > 1 and mz[p] - mz[p-left_shift] < max_dist and intsy[p-left_shift] <= current_intsy and intsy[p-left_shift] > target_intsy:
                left_shift += 1
            lx1, lx2 = mz[p-left_shift], mz[p-left_shift+1]
            ly1, ly2 = intsy[p-left_shift], intsy[p-left_shift+1]
            if not ly1 <= target_intsy <= ly2:
                # warn('Failed to find the left boundary of the peak at %f (probably found an overlapping peak)' % current_mz)
                continue
            # Interpolate the mz values actually corresponding to peak_height_fraction*current_intsy:
            lx = (target_intsy-ly1)*(lx2-lx1)/(ly2-ly1) + lx1
            if not lx1 <= lx <= lx2:
                raise RuntimeError('Failed to interpolate the left boundary mz value of the peak at %f' % current_mz)
            rx = (target_intsy-ry1)*(rx2-rx1)/(ry2-ry1) + rx1
            if not rx1 <= rx <= rx2:
                raise RuntimeError('Failed to interpolate the right boundary mz value of the peak at %f' % current_mz)
            # Join the interpolated boundary with the actual measurements:
            x = np.hstack((lx, mz[(p-left_shift+1):(p+right_shift)], rx))
            y = np.hstack((target_intsy, intsy[(p-left_shift+1):(p+right_shift)], target_intsy))
            # Integrate the area:
            cint = np.trapz(y, x)
            cmz = np.trapz(y*x, x)/cint
            if cmz not in centroid_mz:  # intensity errors may introduce artificial peaks
                centroid_mz.append(cmz)
                centroid_intensity.append(cint)
                # Store the apex data:
                peak_mz.append(current_mz)
                peak_intensity.append(current_intsy)
        return(list(zip(centroid_mz, centroid_intensity)), list(zip(peak_mz, peak_intensity)))

    def resample(self, target_mz, mz_distance_threshold=0.05):
        """
        Returns a resampled spectrum with intensity values approximated
        at points given by a sorted iterable target_mz.
        The approximation is performed by a piecewise linear interpolation
        of the spectrum intensities. The spectrum needs to be in profile mode
        in order for this procedure to work properly.
        The spectrum is interpolated only if two target mz values closest to a
        given target mz are closer than the specified threshold
        This is done in order to interpolate the intensity only within peaks, not between them.
        If the surrounding mz values are further away than the threshold,
        it is assumed that the given target mz corresponds to the background and
        there is no intensity at that point.
        A rule-of-thumb is to set threshold as twice the distance between
        neighboring m/z measurements.
        Large thresholds may lead to non-zero resampled intensity in the background,
        low thresholds might cause bad interpolation due to missing intensity values.
        """
        mz = [mz for mz, intsy in self.confs]
        intsy = [intsy for mz, intsy in self.confs]
        x = target_mz[0]
        for m in target_mz:
            assert m >= x, "The target_mz list is not sorted!"
            x = m
        lenx = len(target_mz)
        lent = len(mz)
        qi = 0  # query (x) index
        ti = 0  # target index - the first index s.t. mz[ti] >= x[qi]
        y = [0.]*lenx  # resampled intensities
        y0, y1 = intsy[0], intsy[0]  # intensities of target spectrum around the point target_mz[qi]
        x0, x1 = mz[0], mz[0]  # mz around the point target_mz[qi]
        # before mz starts, the intensity is zero:
        while target_mz[qi] < mz[0]:
            qi += 1
        # interpolating:
        while ti < lent-1:
            ti += 1
            y0 = y1
            y1 = intsy[ti]
            x0 = x1
            x1 = mz[ti]
            while qi < lenx and target_mz[qi] <= mz[ti]:
                # note: maybe in this case set one of the values to zero to get a better interpolation of edges
                if x1-x0 < mz_distance_threshold:
                    y[qi] = y1 + (target_mz[qi]-x1)*(y0-y1)/(x0-x1)
                qi += 1
        return Spectrum(confs = list(zip(target_mz, y)))


    def fuzzify_peaks(self, sd, step):
        """
        LEGACY FUNCTION. USE SELF.GAUSSIAN_SMOOTING INSTEAD.   
        Applies a gaussian filter to the peaks, effectively broadening them
        and simulating low resolution. Works in place, modifying self.
        The parameter step gives the distance between samples in m/z axis.
        After the filtering, the area below curve (not the sum of intensities!)
        is equal to the sum of the input peak intensities.
        """
        new_mass = np.arange(self.confs[0][0] - 4*sd, self.confs[-1][0] + 4*sd, step)
        A = new_mass[:,np.newaxis] - np.array([m for m,i in self.confs])
        # we don't need to evaluate gaussians to far from their mean,
        # from our perspective 4 standard deviations from the mean is the same
        # as the infinity; this allows to avoid overflow as well:
        A[np.abs(A) > 4*sd] = np.inf
        A **= 2
        A /= (-2*sd**2)
        A = np.exp(A)
        new_intensity = A @ np.array([i for m,i in self.confs])  # matrix multiplication
        new_intensity /= (np.sqrt(2*np.pi)*sd)
        self.set_confs(list(zip(new_mass, new_intensity)))


    def gaussian_smoothing(self, sd=0.01, new_mz=0.01):
        """
        Applies a gaussian filter to the mass spectrum in order to smooth
        it out and decrease the electronic noise.
        Technically, each intensity measurement is replaced by a Gaussian weighted average
        of the neighbouring intensities.  
        As a consequence, the resolution gets decreased.
        Parameter sd (float) controls the width of the gaussian filter.
        Parameter new_mz (float or np.array) is the mass axis of the resulting smoothed spectrum.
        Setting it to float generates an equally spaced mass axis with new_mz being the step length.
        Setting it to np.array sets it as the resulting mass axis.  
        Note that after filtering, the area below curve (not the sum of intensities!)
        is equal to the area of the original spectrum in profile mode,
        or the sum of the input peak intensities in centroid mode.
        """
        if isinstance(new_mz, float):
            new_mz = np.arange(self.confs[0][0] - 4*sd, self.confs[-1][0] + 4*sd, new_mz)
        assert np.all(new_mz[1:] >= new_mz[:-1]), 'The new mz axis needs to be sorted!'
        smooth_intensity = np.zeros(new_mz.shape)
        for mz, intsy in self.confs:
            # smooth_intensity += intsy*np.exp(-(mz - new_mz)**2)**(1/(2*sd**2))
            lpid, rpid = np.searchsorted(new_mz, (mz - 4*sd, mz + 4*sd))
            peak_mz = new_mz[lpid:rpid]
            smooth_intensity[lpid:rpid] += intsy*np.exp(-(mz - peak_mz)**2)**(1/(2*sd**2))
        smooth_intensity /= np.sqrt(2*np.pi)*sd
        self.set_confs(list(zip(new_mz, smooth_intensity)))


    def cut_smallest_peaks(self, removed_proportion=0.001):
        """
        Removes smallest peaks until the total removed intensity amounts
        to the given proportion of the total ion current in the spectrum.
        """
        self.confs.sort(key = lambda x: x[1], reverse=True)
        threshold  = removed_proportion*sum(x[1] for x in self.confs)
        removed = 0
        while len(self.confs)>0 and removed + self.confs[-1][1] <= threshold:
            removed += self.confs.pop()[1]
        self.confs.sort(key = lambda x: x[0])

    def filter_peaks(self, list_of_others, margin):
        """
        Removes peaks which do not match any isotopic envelope from
        the list_of_others, with a given mass margin for matching.
        Works in situ (modifies self).
        Assumes that list_of_others contains proper Spectrum objects
        (i.e. with default sorting of confs).
        _____
        Parameters:
            list_of_others: list
                A list of Spectrum objects.
            margin: float
                The isotopic envelopes of target spectra are widened by this margin.
        _____
        Returns: None
        """
        bounds = [(s.confs[0][0] - margin, s.confs[-1][0] + margin) for s in list_of_others]
        bounds.sort(key = lambda x: x[0])  # sort by lower bound
        merged_bounds = []
        c_low, c_up = bounds[0]
        for b in bounds:
            if b[0] <= c:
                pass # to be finished


    def filter_against_other(self, others, margin=0.15):
        """
        Remove signal from the spectrum which is far from other spectra signal.

        This method removes peaks from self spectrum which m/z is outside the
        area of any peak of other spectra +/- margin. The method does not
        modify self spectrum and returns a new instance of the filtered
        spectrum.

        Parameters
        ----------
        self
            Spectrum to be filtered.
        others:
            One instance of the spectrum against self is filtered or iterable of
            instances of other spectra.
        margin
            m/z radius within signal should be left.

        Returns
        -------
        Spectrum
            A new spectrum with filtered out peaks.

        """
        try:
            other_confs = []
            for other_spectrum in others:
                other_confs.extend(other_spectrum.confs)
            other = Spectrum(confs=other_confs)
        except TypeError:
            other = others
        other_masses = [i[0] for i in other.confs]


        result_confs = []
        index = 0
        for mz, abund in self.confs:
            while (index + 1 < len(other_masses) and
                   other_masses[index + 1] < mz):
                index += 1
            if abs(mz - other_masses[index]) <= margin or (
                    index + 1 < len(other_masses) and
                    abs(mz - other_masses[index + 1]) <= margin):
                result_confs.append((mz, abund))

        result_spectrum = Spectrum(confs=result_confs, label=self.label)
        return result_spectrum

    def plot(self, show = True, profile=False, linewidth=1, **plot_kwargs):
        """
        Plots the spectrum.
        The keyword argument show is retained for backwards compatibility.
        """
        import matplotlib.pyplot as plt
        if profile:
            plt.plot([x[0] for x in self.confs], [x[1] for x in self.confs],
                     linestyle='-', linewidth=linewidth, label=self.label, **plot_kwargs)
        else:
            plt.vlines([x[0] for x in self.confs], [0],
                       [x[1] for x in self.confs], label = self.label,
                       linewidth=linewidth, **plot_kwargs)
        if show:
            plt.show()

    @staticmethod
    def plot_all(spectra, show=True, profile=False, cmap=None, **plot_kwargs):
        """
        Shows the supplied list of spectra on a single plot.
        """
        import matplotlib.pyplot as plt
        import matplotlib.cm as cm
        if not cmap:
            colors = cm.rainbow(np.linspace(0, 1, len(spectra)))
            colors =  [[0, 0, 0, 0.8]] + [list(x[:3]) + [0.6] for x in colors]
        else:
            try:
                colors = [[0, 0, 0, 0.8]] + [cmap(x, alpha=1) for x in range(len(spectra))]
            except:
                colors = cmap
        i = 0
        for spectre in spectra:
            spectre.plot(show=False, profile=profile, color = colors[i],
                         **plot_kwargs)
            i += 1
        #plt.legend(loc=9, bbox_to_anchor=(0.5, -0.1), ncol=len(spectra))  # legend below plot
        plt.legend(loc=0, ncol=1)
        if show:
            plt.show()


if __name__=="__main__":
    import matplotlib.pyplot as plt
    from copy import deepcopy
    S = Spectrum(formula="C2H5OH", threshold=0.01)

    S.add_chemical_noise(4, 0.2)
    S.plot()

    sd = 0.01
    C = deepcopy(S)
    C = C*(np.sqrt(2*np.pi)*sd)**-1
    S1 = deepcopy(S)
    S.gaussian_smoothing(0.01,  0.001)
    S1.fuzzify_peaks(0.01, 0.001)
    S.plot(profile=True, show=False)
    S1.plot(profile=True, show=False)
    C.plot(show=False)
    plt.show()

    T = Spectrum(confs=[(1, 1)])
    T.gaussian_smoothing(0.01, np.array([0.96, 0.97, 0.98, 0.99, 1, 1.01, 1.02]))
    T.plot(profile=True, show=False)
    T = Spectrum(confs=[(1, 1)])
    T.gaussian_smoothing(0.01, 0.001)
    T.plot(profile=True)

    target_mz = np.linspace(45, 56, num=100)
    R = S.resample(target_mz)
    plt.subplot(221)
    S.plot(show=False, profile=True)
    plt.subplot(222)
    R.plot(show=False, profile=True)
    plt.subplot(223)
    S.plot(show=False, profile=True)
    plt.plot([mz for mz, intsy in S.confs],
                   [intsy for mz, intsy in S.confs],
                   'r.')
    plt.subplot(224)
    R.plot(show=False, profile=True)
    plt.plot([mz for mz, intsy in R.confs],
                   [intsy for mz, intsy in R.confs],
                   'r.')
    plt.tight_layout()
    plt.show()



from bisect import bisect_left, bisect_right


def extract_range(L, range_min, range_max):
    '''
    Returns a list of objects l from list L s.t. range_min <= l[0] <= range_max.

    L *must* be sorted.
    '''
    range_min = (range_min, 0.0)
    range_max = (range_max, 0.0)
    left_idx = bisect_left(L, range_min)
    if left_idx == len(L):
        return

    right_idx = bisect_right(L, range_max)

    if right_idx == 0:
        return

    for ii in range(left_idx, right_idx):
        yield L[ii]

def closest(L, x):
    '''
    Returns the element l of L which minimizes abs(l[0], x).

    L *must* be sorted and nonempty.
    '''
    x = (x, 0.0)

    ii = bisect_right(L, x)

    candidates = []
    try:
        candidates = [L[ii]]
    except IndexError:
        pass

    if ii > 0:
        candidates.append(L[ii-1])

    x = x[0]

    return min(candidates, key = lambda tpl: abs(tpl[0] - x))



import numpy as np
from time import time
from masserstein import Spectrum
import pulp as lp
from warnings import warn
import tempfile
from tqdm import tqdm
from pulp.apis import LpSolverDefault
from masserstein import misc



def intensity_generator(confs, mzaxis):
        """
        Generates intensities from spectrum represented as a confs list,
        over m/z values from mzaxis.
        Assumes mzaxis and confs are sorted and returns consecutive intensities.
        """
        mzaxis_id = 0
        mzaxis_len = len(mzaxis)
        for mz, intsy in confs:
            while mzaxis[mzaxis_id] < mz:
                yield 0.
                mzaxis_id += 1
                if mzaxis_id == mzaxis_len:
                    return
            if mzaxis[mzaxis_id] == mz:
                yield intsy
                mzaxis_id += 1
                if mzaxis_id == mzaxis_len:
                    return
        for i in range(mzaxis_id, mzaxis_len):
                yield 0.


def dualdeconv2(exp_sp, thr_sps, penalty, quiet=True, solver=LpSolverDefault):
        """
        This function solves linear program describing optimal transport of signal between the experimental spectrum
        and the list of theoretical (reference) spectra. Additionally, an auxiliary point is introduced in order to
        remove noise from the experimental spectrum, as described by Ciach et al., 2020. 

        _____
        Parameters:
            exp_sp: Spectrum object
                Experimental spectrum.
            thr_sp: list of Spectrum objects
                List of theoretical (reference) spectra.
            penalty: float
                Denoising penalty.
            solver: 
                Which solver should be used. In case of problems with the default solver,
                pulp.GUROBI() is recommended (note that it requires obtaining a licence).
                To see all solvers available at your machine execute: pulp.listSolvers(onlyAvailable=True).

        _____
        Returns: dict
            Dictionary with the following entries:

            - probs: List containing proportions of consecutive theoretical (reference) spectra in the experimental
            spectrum. Note that they do not have to sum up to 1, because some part of the signal can be noise.

            - trash: Amount of noise in the consecutive m/z points of the experimental spectrum.

            - fun: Optimal value of the objective function.

            - status: Status of the linear program.

        """
        start = time()
        exp_confs = exp_sp.confs.copy()
        thr_confs = [thr_sp.confs.copy() for thr_sp in thr_sps]

        # Normalization check:
        assert np.isclose(sum(x[1] for x in exp_confs) , 1), 'Experimental spectrum not normalized'
        for i, thrcnf in enumerate(thr_confs):
                assert np.isclose(sum(x[1] for x in thrcnf), 1), 'Theoretical spectrum %i not normalized' % i

        # Computing a common mass axis for all spectra
        exp_confs = [(m, i) for m, i in exp_confs]
        thr_confs = [[(m, i) for m, i in cfs] for cfs in thr_confs]
        global_mass_axis = set(x[0] for x in exp_confs)
        global_mass_axis.update(x[0] for s in thr_confs for x in s)
        global_mass_axis = sorted(global_mass_axis)
        if not quiet:
                print("Global mass axis computed")
        n = len(global_mass_axis)
        k = len(thr_confs)

        # Computing lengths of intervals between mz measurements (l_i variables)
        interval_lengths = [global_mass_axis[i+1] - global_mass_axis[i] for i in range(n-1)]
        if not quiet:
                print("Interval lengths computed")

        # linear program:
        program = lp.LpProblem('Dual_L1_regression_sparse', lp.LpMaximize)
        if not quiet:
                print("Linear program initialized")
        # variables:
        lpVars = []
        for i in range(n):
                lpVars.append(lp.LpVariable('Z%i' % (i+1), None, penalty, lp.LpContinuous))
        ##        # in case one would like to explicitly forbid non-experimental abyss:
        ##        if V[i] > 0:
        ##            lpVars.append(lp.LpVariable('W%i' % (i+1), None, penalty, lp.LpContinuous))
        ##        else:
        ##            lpVars.append(lp.LpVariable('W%i' % (i+1), None, None, lp.LpContinuous))
        if not quiet:
                print("Variables created")
        # objective function:
        exp_vec = intensity_generator(exp_confs, global_mass_axis)  # generator of experimental intensity observations
        program += lp.lpSum(v*x for v, x in zip(exp_vec, lpVars)), 'Dual_objective'
        # constraints:
        for j in range(k):
                thr_vec = intensity_generator(thr_confs[j], global_mass_axis)
                program += lp.lpSum(v*x for v, x in zip(thr_vec, lpVars) if v > 0.) <= 0, 'P%i' % (j+1)
        if not quiet:
                print('tsk tsk')
        ##    for i in range(n-1):
        ##        program += lpVars[i]-lpVars[i+1] <= interval_lengths[i], 'EpsPlus %i' % (i+1)
        ##        program += lpVars[i] - lpVars[i+1] >=  -interval_lengths[i], 'EpsMinus %i' % (i+1)
        for i in range(n-1):
                program +=  lpVars[i] - lpVars[i+1]  <=  interval_lengths[i], 'EpsPlus_%i' % (i+1)
                program +=  lpVars[i] - lpVars[i+1]  >= -interval_lengths[i], 'EpsMinus_%i' % (i+1)
        if not quiet:
                print("Constraints written")
        #program.writeLP('WassersteinL1.lp')
        if not quiet:
                print("Starting solver")
        LpSolverDefault.msg = not quiet
        program.solve(solver = solver)
        end = time()
        if not quiet:
                print("Solver finished.")
                print("Status:", lp.LpStatus[program.status])
                print("Optimal value:", lp.value(program.objective))
                print("Time:", end - start)
        constraints = program.constraints
        probs = [round(constraints['P%i' % i].pi, 12) for i in range(1, k+1)]
        exp_vec = list(intensity_generator(exp_confs, global_mass_axis))
        # 'if' clause below is to restrict returned abyss to experimental confs
        abyss = [round(x.dj, 12) for i, x in enumerate(lpVars) if exp_vec[i] > 0.]
        # note: accounting for number of summands in checking of result correctness,
        # because summation of many small numbers introduces numerical errors
        if not np.isclose(sum(probs)+sum(abyss), 1., atol=len(abyss)*1e-03):
                warn("""In dualdeconv2:
                Proportions of signal and noise sum to %f instead of 1.
                This may indicate improper results.
                Please check the deconvolution results and consider reporting this warning to the authors.
                                    """ % (sum(probs)+sum(abyss)))

        return {"probs": probs, "trash": abyss, "fun": lp.value(program.objective), 'status': program.status}


def dualdeconv2_alternative(exp_sp, thr_sps, penalty, quiet=True, solver=LpSolverDefault):

        """
        Alternative version of dualdeconv2 - using .pi instead of .dj to extract optimal values of variables.
        Slower, but .pi is better documented than .dj in pulp. Gives the same results as dualdeconv2.
        This function solves linear program describing optimal transport of signal between the experimental spectrum
        and the list of theoretical (reference) spectra. Additionally, an auxiliary point is introduced in order to
        remove noise from the experimental spectrum, as described by Ciach et al., 2020. 

        _____
        Parameters:
            exp_sp: Spectrum object
                Experimental spectrum.
            thr_sp: list of Spectrum objects
                List of theoretical (reference) spectra.
            penalty: float
                Denoising penalty.
            solver: 
                Which solver should be used. In case of problems with the default solver,
                pulp.GUROBI() is recommended (note that it requires obtaining a licence).
                To see all solvers available at your machine execute: pulp.listSolvers(onlyAvailable=True).

        _____
        Returns: dict
            Dictionary with the following entries:

            - probs: List containing proportions of consecutive theoretical (reference) spectra in the experimental
            spectrum. Note that they do not have to sum up to 1, because some part of the signal can be noise.

            - trash: Amount of noise in the consecutive m/z points of the experimental spectrum.

            - fun: Optimal value of the objective function.

            - status: Status of the linear program.

        """

        start = time()
        exp_confs = exp_sp.confs.copy()
        thr_confs = [thr_sp.confs.copy() for thr_sp in thr_sps]

        # Normalization check:
        assert np.isclose(sum(x[1] for x in exp_confs) , 1), 'Experimental spectrum not normalized'
        for i, thrcnf in enumerate(thr_confs):
                assert np.isclose(sum(x[1] for x in thrcnf), 1), 'Theoretical spectrum %i not normalized' % i

        # Computing a common mass axis for all spectra
        exp_confs = [(m, i) for m, i in exp_confs]
        thr_confs = [[(m, i) for m, i in cfs] for cfs in thr_confs]
        global_mass_axis = set(x[0] for x in exp_confs)
        global_mass_axis.update(x[0] for s in thr_confs for x in s)
        global_mass_axis = sorted(global_mass_axis)
        if not quiet:
                print("Global mass axis computed")
        n = len(global_mass_axis)
        k = len(thr_confs)

        # Computing lengths of intervals between mz measurements (l_i variables)
        interval_lengths = [global_mass_axis[i+1] - global_mass_axis[i] for i in range(n-1)]
        if not quiet:
                print("Interval lengths computed")

        # linear program:
        program = lp.LpProblem('Dual_L1_regression_sparse', lp.LpMaximize)
        if not quiet:
                print("Linear program initialized")
        # variables:
        lpVars = []
        for i in range(n):
                lpVars.append(lp.LpVariable('Z%i' % (i+1), None, None, lp.LpContinuous))
        ##        # in case one would like to explicitly forbid non-experimental abyss:
        ##        if V[i] > 0:
        ##            lpVars.append(lp.LpVariable('W%i' % (i+1), None, penalty, lp.LpContinuous))
        ##        else:
        ##            lpVars.append(lp.LpVariable('W%i' % (i+1), None, None, lp.LpContinuous))
        if not quiet:
                print("Variables created")
        # objective function:
        exp_vec = intensity_generator(exp_confs, global_mass_axis)  # generator of experimental intensity observations
        program += lp.lpSum(v*x for v, x in zip(exp_vec, lpVars)), 'Dual_objective'
        # constraints:
        for j in range(k):
                thr_vec = intensity_generator(thr_confs[j], global_mass_axis)
                program += lp.lpSum(v*x for v, x in zip(thr_vec, lpVars) if v > 0.) <= 0, 'P%i' % (j+1)
        if not quiet:
                print('tsk tsk')
        for i in range(n):
                program += lpVars[i] <= penalty, 'g%i' % (i+1)
        for i in range(n-1):
                program +=  lpVars[i] - lpVars[i+1]  <=  interval_lengths[i], 'EpsPlus_%i' % (i+1)
                program +=  lpVars[i] - lpVars[i+1]  >= -interval_lengths[i], 'EpsMinus_%i' % (i+1)
        if not quiet:
                print("Constraints written")
        #program.writeLP('WassersteinL1.lp')
        if not quiet:
                print("Starting solver")
        LpSolverDefault.msg = not quiet
        program.solve(solver = solver)
        end = time()
        if not quiet:
                print("Solver finished.")
                print("Status:", lp.LpStatus[program.status])
                print("Optimal value:", lp.value(program.objective))
                print("Time:", end - start)
        constraints = program.constraints
        probs = [round(constraints['P%i' % i].pi, 12) for i in range(1, k+1)]
        exp_vec = list(intensity_generator(exp_confs, global_mass_axis))
        abyss = [round(constraints['g%i' % i].pi, 12) for i in range(1, n+1) if exp_vec[i-1] > 0.]
        # note: accounting for number of summands in checking of result correctness,
        # because summation of many small numbers introduces numerical errors
        if not np.isclose(sum(probs)+sum(abyss), 1., atol=len(abyss)*1e-03):
                warn("""In dualdeconv2_alternative:
                Proportions of signal and noise sum to %f instead of 1.
                This may indicate improper results.
                Please check the deconvolution results and consider reporting this warning to the authors.
                                    """ % (sum(probs)+sum(abyss)))

        return {"probs": probs, "trash": abyss, "fun": lp.value(program.objective), 'status': program.status}



def dualdeconv3(exp_sp, thr_sps, penalty, penalty_th, quiet=True, solver=LpSolverDefault):

        """
        This function solves linear program describing optimal transport of signal between 
        the experimental spectrum and the list of theoretical (reference) spectra. 
        Two auxiliary points are introduced in order to remove noise from the experimental spectrum
        and from the combination of theoretical (reference) spectra. 
        Transport of signal between the two auxiliary points is explicitly forbidden.
        Mathematically, this formulation is equivalent to the one implemented in dualdeconv4
        and both give the same results up to roundoff errors.

        _____
        Parameters:
            exp_sp: Spectrum object
                Experimental spectrum.
            thr_sp: Spectrum object
                List of theoretical (reference) spectra.
            penalty: float
                Denoising penalty for the experimental spectrum.
            penalty_th: float
                Denoising penalty for the theoretical (reference) spectra.
            solver: 
                Which solver should be used. In case of problems with the default solver,
                pulp.GUROBI() is recommended (note that it requires obtaining a licence).
                To see all solvers available at your machine execute: pulp.listSolvers(onlyAvailable=True).

        _____
        Returns: dict
            Dictionary with the following entries:

            - probs: List containing proportions of consecutive theoretical (reference) spectra in the experimental
            spectrum. Note that they do not have to sum up to 1, because some part of the signal can be noise.

            - noise_in_theoretical: Proportion of noise present in the combination of theoretical
            (reference) spectra.

            - trash: Amount of noise in the consecutive m/z points of the experimental spectrum.

            - theoretical trash: Amount of noise present in the combination of theoretical (reference)
            spectra in consecutive m/z points from global mass axis.

            - fun: Optimal value of the objective function.

            - status: Status of the linear program.

            - global mass axis: All the m/z values from the experimental spectrum and from the theoretical 
            (reference) spectra in a sorted list. 

        """

        start = time()
        exp_confs = exp_sp.confs.copy()
        thr_confs = [thr_sp.confs.copy() for thr_sp in thr_sps]

        # Normalization check:
        assert np.isclose(sum(x[1] for x in exp_confs) , 1), 'Experimental spectrum not normalized'
        for i, thrcnf in enumerate(thr_confs):
                assert np.isclose(sum(x[1] for x in thrcnf), 1), 'Theoretical spectrum %i not normalized' % i

        # Computing a common mass axis for all spectra
        exp_confs = [(m, i) for m, i in exp_confs]
        thr_confs = [[(m, i) for m, i in cfs] for cfs in thr_confs]
        global_mass_axis = set(x[0] for x in exp_confs)
        global_mass_axis.update(x[0] for s in thr_confs for x in s)
        global_mass_axis = sorted(global_mass_axis)
        if not quiet:
                print("Global mass axis computed")
        n = len(global_mass_axis)
        k = len(thr_confs)

        # Computing lengths of intervals between mz measurements (l_i variables)
        interval_lengths = [global_mass_axis[i+1] - global_mass_axis[i] for i in range(n-1)]
        if not quiet:
                print("Interval lengths computed")

        # linear program:
        program = lp.LpProblem('Dual_L1_regression_sparse', lp.LpMaximize)
        if not quiet:
                print("Linear program initialized")

        # variables:
        lpVars = []
        try:
                for i in range(n-2):
                        lpVars.append(lp.LpVariable('Z%i' % (i+1), None, None, lp.LpContinuous))
                lpVars.append(lp.LpVariable('Z%i' % (n-1), -interval_lengths[n-2], interval_lengths[n-2], lp.LpContinuous))
        except IndexError:
                pass
        lpVars.append(lp.LpVariable('Z%i' % (n), None, None, lp.LpContinuous))
        lpVars.append(lp.LpVariable('Z%i' % (n+1), None, None, lp.LpContinuous))
        lpVars.append(lp.LpVariable('Z%i' % (n+2), 0, None, lp.LpContinuous))
        lpVars.append(lp.LpVariable('Z%i' % (n+3), 0, None, lp.LpContinuous))

        if not quiet:
                print("Variables created")

        # objective function:
        exp_vec = intensity_generator(exp_confs, global_mass_axis)  # generator of experimental intensity observations
        program += lp.lpSum(v*x for v, x in zip(exp_vec, lpVars[:n-1]+[0])).addInPlace(
                                lp.lpSum(v*x for v, x in zip([-1, 0, 0, -1], lpVars[n-1:]))), 'Dual_objective'

        # constraints:
        for j in range(k):
                thr_vec = intensity_generator(thr_confs[j], global_mass_axis)
                program += lp.lpSum(v*x for v, x in zip(thr_vec, lpVars[:n-1]+[0]) if v > 0.).addInPlace(
                                lp.lpSum(v*x for v, x in zip([-1, 0, 1, -1], lpVars[n-1:]))) <= 0, 'P_%i' % (j+1)

        exp_vec = intensity_generator(exp_confs, global_mass_axis)
        program += lp.lpSum(v*x for v, x in zip(exp_vec, lpVars[:n-1]+[0])).addInPlace(
                                lp.lpSum(v*x for v, x in zip([0, 1, -1, 0], lpVars[n-1:]))) <= 0, 'p0_prime'

        if not quiet:
                print('tsk tsk')

        for i in range(n-1):
                program +=  lpVars[i] - lpVars[n-1]  <=  penalty, 'g_%i' % (i+1)
                program +=  -lpVars[n] - lpVars[i] <= penalty_th, 'g_prime_%i' % (i+1)
        try:
                for i in range(n-2):
                        program += lpVars[i] - lpVars[i+1] <= interval_lengths[i], 'epsilon_plus_%i' % (i+1)
                        program += lpVars[i+1] - lpVars[i] <= interval_lengths[i], 'epsilon_minus_%i' % (i+1)
        except IndexError:
                pass

        program += -lpVars[n-1] <= penalty, 'g_%i' % (n)
        program += -lpVars[n] <= penalty_th, 'g_prime_%i' % (n)

        if not quiet:
                print("Constraints written")

        if not quiet:
                print("Starting solver")

        #Solving
        LpSolverDefault.msg = not quiet
        program.solve(solver = solver)
        end = time()
        if not quiet:
                print("Solver finished.")
                print("Status:", lp.LpStatus[program.status])
                print("Optimal value:", lp.value(program.objective))
                print("Time:", end - start)
        constraints = program.constraints
        probs = [round(constraints['P_%i' % i].pi, 12) for i in range(1, k+1)]
        p0_prime = round(constraints['p0_prime'].pi, 12)
        exp_vec = list(intensity_generator(exp_confs, global_mass_axis))
        abyss = [round(constraints['g_%i' % i].pi, 12) for i in range(1, n+1) if exp_vec[i-1] > 0.]
        abyss_th = [round(constraints['g_prime_%i' % i].pi, 12) for i in range(1, n+1)]

        if not np.isclose(sum(probs)+sum(abyss), 1., atol=len(abyss)*1e-03):
                warn("""In dualdeconv3:
                Proportions of signal and noise sum to %f instead of 1.
                This may indicate improper results.
                Please check the deconvolution results and consider reporting this warning to the authors.
                                    """ % (sum(probs)+sum(abyss)))

        return {"probs": probs, "noise_in_theoretical": p0_prime, "trash": abyss, "theoretical_trash": abyss_th,
         "fun": lp.value(program.objective), 'status': program.status, 'global_mass_axis': global_mass_axis}


def dualdeconv4(exp_sp, thr_sps, penalty, penalty_th, quiet=True, solver=LpSolverDefault):

        """
        This function solves linear program describing optimal transport of signal between the experimental 
        spectrum and the list of theoretical (reference) spectra. 
        Two auxiliary points are introduced in order to remove noise from the experimental spectrum
        and from the combination of theoretical (reference) spectra. 
        Transport of signal between the two auxiliary points is allowed (with cost equal to penalty + penalty_th),
        however, it is not optimal so it never occurs. Mathematically, this formulation is equivalent to the one 
        implemented in dualdeconv3 and both give the same results up to roundoff errors.

        _____
        Parameters:
            exp_sp: Spectrum object
                Experimental spectrum.
            thr_sp: Spectrum object
                List of theoretical (reference) spectra.
            penalty: float
                Denoising penalty for the experimental spectrum.
            penalty_th: float
                Denoising penalty for the theoretical (reference) spectra.
            solver: 
                Which solver should be used. In case of problems with the default solver,
                pulp.GUROBI() is recommended (note that it requires obtaining a licence).
                To see all solvers available at your machine execute: pulp.listSolvers(onlyAvailable=True).
        
        _____
        Returns: dict
            Dictionary with the following entries:

            - probs: List containing proportions of consecutive theoretical (reference) spectra in the experimental
            spectrum. Note that they do not have to sum up to 1, because some part of the signal can be noise.

            - noise_in_theoretical: Proportion of noise present in the combination of theoretical
            (reference) spectra.

            - trash: Amount of noise in the consecutive m/z points of the experimental spectrum.

            - theoretical trash: Amount of noise present in the combination of theoretical (reference)
            spectra in consecutive m/z points from global mass axis.

            - fun: Optimal value of the objective function.

            - status: Status of the linear program.

            - global mass axis: All the m/z values from the experimental spectrum and from the theoretical 
            (reference) spectra in a sorted list. 

        """

        start = time()
        exp_confs = exp_sp.confs.copy()
        thr_confs = [thr_sp.confs.copy() for thr_sp in thr_sps]

        # Normalization check:
        assert np.isclose(sum(x[1] for x in exp_confs) , 1), 'Experimental spectrum not normalized'
        for i, thrcnf in enumerate(thr_confs):
                assert np.isclose(sum(x[1] for x in thrcnf), 1), 'Theoretical spectrum %i not normalized' % i

        # Computing a common mass axis for all spectra
        exp_confs = [(m, i) for m, i in exp_confs]
        thr_confs = [[(m, i) for m, i in cfs] for cfs in thr_confs]
        global_mass_axis = set(x[0] for x in exp_confs)
        global_mass_axis.update(x[0] for s in thr_confs for x in s)
        global_mass_axis = sorted(global_mass_axis)
        if not quiet:
                print("Global mass axis computed")
        n = len(global_mass_axis)
        k = len(thr_confs)

        # Computing lengths of intervals between mz measurements (l_i variables)
        interval_lengths = [global_mass_axis[i+1] - global_mass_axis[i] for i in range(n-1)]
        if not quiet:
                print("Interval lengths computed")

        # linear program:
        program = lp.LpProblem('Dual_L1_regression_sparse', lp.LpMaximize)
        if not quiet:
                print("Linear program initialized")

        # variables:
        lpVars = []
        try:
                for i in range(n-2):
                        lpVars.append(lp.LpVariable('Z%i' % (i+1), None, None, lp.LpContinuous))
                lpVars.append(lp.LpVariable('Z%i' % (n-1), -interval_lengths[n-2], interval_lengths[n-2], lp.LpContinuous))
        except IndexError:
                pass
        lpVars.append(lp.LpVariable('Z%i' % n, None, None, lp.LpContinuous))
        lpVars.append(lp.LpVariable('Z%i' % (n+1), None, None, lp.LpContinuous))
        lpVars.append(lp.LpVariable('Z%i' % (n+2), 0, None, lp.LpContinuous))
        lpVars.append(lp.LpVariable('Z%i' % (n+3), 0, None, lp.LpContinuous))
        if not quiet:
                print("Variables created")

        # objective function:
        exp_vec = intensity_generator(exp_confs, global_mass_axis)  # generator of experimental intensity observations
        program += lp.lpSum(v*x for v, x in zip(exp_vec, lpVars[:n-1]+[0])).addInPlace(
                            lp.lpSum(v*x for v, x in zip([-1, 0, 0, -1], lpVars[n-1:]))), 'Dual_objective'

        # constraints:
        for j in range(k):
                thr_vec = intensity_generator(thr_confs[j], global_mass_axis)
                program += lp.lpSum(v*x for v, x in zip(thr_vec, lpVars[:n-1]+[0]) if v > 0.).addInPlace(
                            lp.lpSum(v*x for v, x in zip([-1, 0, 1, -1], lpVars[n-1:]))) <= -penalty, 'P_%i' % (j+1)

        exp_vec = intensity_generator(exp_confs, global_mass_axis)
        program += lp.lpSum(v*x for v, x in zip(exp_vec, lpVars[:n-1]+[0])).addInPlace(
                            lp.lpSum(v*x for v, x in zip([0, 1, -1, 0], lpVars[n-1:]))) <= penalty_th, 'p0_prime'

        if not quiet:
                print('tsk tsk')
        
        for i in range(n-1):
                program +=  lpVars[i] - lpVars[n-1]  <=  0, 'g_%i' % (i+1)
                program +=  -lpVars[i] - lpVars[n]  <= 0, 'g_prime_%i' % (i+1)
        try:
                for i in range(n-2):
                        program += lpVars[i] - lpVars[i+1] <= interval_lengths[i], 'epsilon_plus_%i' % (i+1)
                        program += lpVars[i+1] - lpVars[i] <= interval_lengths[i], 'epsilon_minus_%i' % (i+1)
        except IndexError:
                pass

        program += -lpVars[n-1] <= 0, 'g_%i' % (n)
        program += -lpVars[n] <= 0, 'g_prime_%i' % (n)

        if not quiet:
                print("Constraints written")

        if not quiet:
                print("Starting solver")

        #Solving
        LpSolverDefault.msg = not quiet
        program.solve(solver = solver)
        end = time()
        if not quiet:
                print("Solver finished.")
                print("Status:", lp.LpStatus[program.status])
                print("Optimal value:", lp.value(program.objective))
                print("Time:", end - start)
        constraints = program.constraints
        probs = [round(constraints['P_%i' % i].pi, 12) for i in range(1, k+1)]
        p0_prime = round(constraints['p0_prime'].pi, 12)
        exp_vec = list(intensity_generator(exp_confs, global_mass_axis))
        abyss = [round(constraints['g_%i' % i].pi, 12) for i in range(1, n+1) if exp_vec[i-1] > 0.]
        abyss_th = [round(constraints['g_prime_%i' % i].pi, 12) for i in range(1, n+1)]
        if not np.isclose(sum(probs)+sum(abyss), 1., atol=len(abyss)*1e-03):
                warn("""In dualdeconv4:
                Proportions of signal and noise sum to %f instead of 1.
                This may indicate improper results.
                Please check the deconvolution results and consider reporting this warning to the authors.
                                    """ % (sum(probs)+sum(abyss)))

        return {"probs": probs, "noise_in_theoretical": p0_prime, "trash": abyss, "theoretical_trash": abyss_th, 
        "fun": lp.value(program.objective)+penalty, 'status': program.status, 'global_mass_axis': global_mass_axis}


def estimate_proportions(spectrum, query, MTD=0.1, MDC=1e-8,
                        MMD=-1, max_reruns=3, verbose=False, 
                        progress=True, MTD_th=0.1, solver=LpSolverDefault):

    """
    Returns estimated proportions of molecules from query in spectrum.
    Performs initial filtering of formulas and experimental spectrum to speed up the computations.

    _____
    Parameters:
    spectrum: Spectrum object
        The experimental (subject) spectrum.
    query: list of Spectrum objects
        A list of theoretical (reference) spectra.
    MTD: Maximum Transport Distance, float
        Ion current from experimental spectrum will be transported up to this distance when estimating
        molecule proportions. Default is 1.
    MDC: Minimum Detectable Current, float
        If the isotopic envelope of an ion encompasses less than
        this amount of the total ion current, it is assumed that this ion
        is absent in the spectrum. Default is 1e-8.
    MMD: Maximum Mode Distance, float
        If there is no experimental peak within this distance from the
        highest peak of an isotopic envelope of a molecule,
        it is assumed that this molecule is absent in the spectrum.
        Setting this value to -1 disables filtering. Default is -1.
    max_reruns: int
        Due to numerical errors, some partial results may be inaccurate.
        If this is detected, then those results are recomputed for a maximal number of times
        given by this parameter. Default is 3.
    verbose: bool
        Print diagnistic messages? Default is False.
    progress: bool
        Whether to display progress bars during work. Default is True.
    MTD_th: Maximum Transport Distance for theoretical spectra, float
        If presence of noise in theoretical (reference, query) spectra is not expected, 
        then this parameter should be set to None. Otherwise, set its value to some positive real number.
        Ion current from theoretical spectra will be transported up to this distance 
        when estimating molecule proportions. Default is None.
    solver: 
        Which solver should be used. In case of problems with the default solver,
        pulp.GUROBI() is recommended (note that it requires obtaining a licence).
        To see all solvers available at your machine execute: pulp.listSolvers(onlyAvailable=True).

    _____
    Returns: dict
        A dictionary with the following entries:

        - proportions: List of proportions of query (i.e. theoretical, reference) spectra.

        - noise: List of intensities that could not be explained by the supplied formulas. 
        The intensities correspond to the m/z values of the experimental spectrum.

        If MTD_th parameter is not equal to None, then the dictionary contains also 
        the following entries:

        - noise_in_theoretical: List of intensities from query (i.e. theoretical, reference) spectra
        that do not correspond to any intensities in the experimental spectrum and therefore were 
        identified as noise. The intensities correspond to the m/z values from global mass axis.

        - proportion_of_noise_in_theoretical: Proportion of noise present in the combination of query
        (i.e. theoretical, reference) spectra.

        - global_mass_axis: All the m/z values from the experimental spectrum and from the query 
        (i.e. theoretical, reference) spectra in a sorted list. 

    """

    def progr_bar(x, **kwargs):
        if progress:
            return tqdm(x, **kwargs)
        else:
            return x
    try:
        exp_confs = spectrum.confs
    except:
        print("Could not retrieve the confs list. Is the supplied spectrum an object of class Spectrum?")
        raise
    assert abs(sum(x[1] for x in exp_confs) - 1.) < 1e-08, 'The experimental spectrum is not normalized.'
    assert all(x[0] >= 0. for x in exp_confs), 'Found experimental peaks with negative masses!'
    if any(x[1] < 0 for x in exp_confs):
        raise ValueError("""
        The experimental spectrum cannot contain negative intensities. 
        Please remove them using e.g. the Spectrum.trim_negative_intensities() method.
        """)
                           
    vortex = [0.]*len(exp_confs)  # unexplained signal
    k = len(query)
    proportions = [0.]*k

    if MTD_th is None:
        MTD_max = MTD
    else:
        MTD_max = max(MTD, MTD_th)

    for i, q in enumerate(query):
        assert abs(sum(x[1] for x in q.confs) - 1.) < 1e-08, 'Theoretical spectrum %i is not normalized' %i
        assert all(x[0] >= 0 for x in q.confs), 'Theoretical spectrum %i has negative masses!' % i

    # Initial filtering of formulas
    envelope_bounds = []
    filtered = []
    for i in progr_bar(range(k), desc = "Initial filtering of formulas"):
        s = query[i]
        mode = s.get_modal_peak()[0]
        mn = s.confs[0][0]
        mx = s.confs[-1][0]
        matching_current = MDC==0. or sum(x[1] for x in misc.extract_range(
                                                                        exp_confs, mn - MTD_max, mx + MTD_max)) >= MDC
        matching_mode = MMD==-1 or abs(misc.closest(exp_confs, mode)[0] - mode) <= MMD

        if matching_mode and matching_current:
            envelope_bounds.append((mn, mx, i))
        else:
            envelope_bounds.append((-1, -1, i))
            filtered.append(i)

    envelope_bounds.sort(key=lambda x: x[0])  # sorting by lower bounds
    if verbose:
        print("Removed theoretical spectra due to no matching experimental peaks:", filtered)
        print('Envelope bounds:', envelope_bounds)

    # Computing chunks
    chunkIDs = [0]*k  # Grouping of theoretical spectra
    # Note: order of chunkIDs corresponds to order of query, not the envelope bounds
    # chunk_bounds = mass intervals matching chunks, accounting for mass transport
    # order of chunk_bounds corresponds to increasing chunk ID,
    # so that chunk_bounds[0] is the interval for chunk nr 0
    chunk_bounds = []
    current_chunk = 0
    first_present = 0
    while envelope_bounds[first_present][0] == -1 and first_present < k-1:
        _, _, sp_id = envelope_bounds[first_present]
        chunkIDs[sp_id] = -1
        first_present += 1
    prev_mn, prev_mx, prev_id = envelope_bounds[first_present]
    for i in progr_bar(range(first_present, k), desc = "Computing chunks"):
        mn, mx, sp_id = envelope_bounds[i]
        if mn - prev_mx > 2*MTD_max:
            current_chunk += 1
            chunk_bounds.append( (prev_mn-MTD_max, prev_mx+MTD_max) )
            prev_mn = mn  # get lower bound of new chunk
        prev_mx = mx  # update the lower bound of current chunk
        chunkIDs[sp_id] = current_chunk
    chunk_bounds.append( (prev_mn-MTD_max, prev_mx+MTD_max) )
    nb_of_chunks = len(chunk_bounds)
    if verbose:
        print('Number of chunks: %i' % nb_of_chunks)
        print("ChunkIDs:", chunkIDs)
        print("Chunk bounds:", chunk_bounds)

    # Splitting the experimental spectrum into chunks
    exp_conf_chunks = []  # list of indices of experimental confs matching chunks
    current_chunk = 0
    matching_confs = []  # experimental confs matching current chunk
    cur_bound = chunk_bounds[current_chunk]
    for conf_id, cur_conf in progr_bar(enumerate(exp_confs), desc = "Splitting the experimental spectrum into chunks"):
        while cur_bound[1] < cur_conf[0] and current_chunk < nb_of_chunks-1:
            exp_conf_chunks.append(matching_confs)
            matching_confs = []
            current_chunk += 1
            cur_bound = chunk_bounds[current_chunk]
        if cur_bound[0] <= cur_conf[0] <= cur_bound[1]:
            matching_confs.append(conf_id)
        else:
            # experimental peaks outside chunks go straight to vortex
            vortex[conf_id] = cur_conf[1]
    exp_conf_chunks.append(matching_confs)
    chunk_TICs = [sum(exp_confs[i][1] for i in chunk_list) for chunk_list in exp_conf_chunks]
    if verbose:
        # print('Trash after filtering:', vortex)
        print("Ion currents in chunks:", chunk_TICs)

    # Deconvolving chunks:
    p0_prime = 0
    vortex_th = []
    global_mass_axis = []
    for current_chunk_ID, conf_IDs in progr_bar(enumerate(exp_conf_chunks), desc="Deconvolving chunks", 
                                                                            total=len(exp_conf_chunks)):
        if verbose:
            print("Deconvolving chunk %i" % current_chunk_ID)
        if chunk_TICs[current_chunk_ID] < 1e-16:
            # nothing to deconvolve, pushing remaining signal to vortex
            if verbose:
                print('Chunk %i is almost empty - skipping deconvolution' % current_chunk_ID)
            for i in conf_IDs:
                vortex[i] = exp_confs[i][1]
        else:
            chunkSp = Spectrum('', empty=True)
            # Note: conf_IDs are monotonic w.r.t. conf mass,
            # so constructing a spectrum will not change the order
            # of confs supplied in the list below:
            chunkSp.set_confs([exp_confs[i] for i in conf_IDs])
            chunkSp.normalize()
            theoretical_spectra_IDs = [i for i, c in enumerate(chunkIDs) if c == current_chunk_ID]
            thrSp = [query[i] for i in theoretical_spectra_IDs]

            rerun = 0
            success = False
            while not success:
                    rerun += 1
                    if rerun > max_reruns:
                            raise RuntimeError('Failed to deconvolve a fragment of the experimental spectrum \
                                                with mass (%f, %f)' % chunk_bounds[current_chunk_ID])
                    if MTD_th is None:
                        dec = dualdeconv2(chunkSp, thrSp, MTD, quiet=True, solver=solver)
                    else:
                        dec = dualdeconv4(chunkSp, thrSp, MTD, MTD_th, quiet=True, solver=solver)
                    if dec['status'] == 1:
                            success=True
                    else:
                            warn('Rerunning computations for chunk %i due to status %s' % (current_chunk_ID, 
                                                                                        lp.LpStatus[dec['status']]))
            if verbose:
                    print('Chunk %i deconvolution status:', lp.LpStatus[dec['status']])
                    print('Signal proportion in experimental spectrum:', sum(dec['probs']))
                    print('Noise proportion in experimental spectrum:', sum(dec['trash']))
                    print('Total explanation:', sum(dec['probs'])+sum(dec['trash']))
                    if MTD_th is not None:
                        print('Noise proportion in combination of theoretical spectra:', dec["noise_in_theoretical"])
            for i, p in enumerate(dec['probs']):
                original_thr_spectrum_ID = theoretical_spectra_IDs[i]
                proportions[original_thr_spectrum_ID] = p*chunk_TICs[current_chunk_ID]
            for i, p in enumerate(dec['trash']):
                original_conf_id = conf_IDs[i]
                vortex[original_conf_id] = p*chunk_TICs[current_chunk_ID]
            if MTD_th is not None:
                p0_prime = p0_prime + dec["noise_in_theoretical"]*chunk_TICs[current_chunk_ID]
                rescaled_vortex_th = [element*chunk_TICs[current_chunk_ID] for element in dec['theoretical_trash']]
                vortex_th = vortex_th + rescaled_vortex_th
                global_mass_axis = global_mass_axis + dec['global_mass_axis']

    if not np.isclose(sum(proportions)+sum(vortex), 1., atol=len(vortex)*1e-03):
        warn("""In estimate_proportions:
Proportions of signal and noise sum to %f instead of 1.
This may indicate improper results.
Please check the deconvolution results and consider reporting this warning to the authors.
                        """ % (sum(proportions)+sum(vortex)))
    if MTD_th is not None:
        return {'proportions': proportions, 'noise': vortex, 'noise_in_theoretical': vortex_th, 
                'proportion_of_noise_in_theoretical': p0_prime, 'global_mass_axis': global_mass_axis}
    else:
        return {'proportions': proportions, 'noise': vortex}


if __name__=="__main__":

    exper = [(1., 1/6.), (2., 3/6.), (3., 2/6.)]
    thr1 = [(1., 1/2.), (2.,1/2.)]
    thr2 = [(2., 1/2.), (3., 1/2.)]

    exper = [(1, 0.25), (3, 0.5), (6, 0.25)]
    thr1 = [(1., 1.), (3, 0.)]
    thr2 = [(3, 0.5), (4, 0.5)]
    thr = [thr1, thr2]

    exper = [(1.1, 1/3), (2.2, 5/12), (3.1, 1/4)]


    exper = [(0, 1/4), (1.1, 1/6), (2.2, 5/24), (3.1, 1/8), (4, 1/4), (60, .1) ]
    ##thr1 = [(1, 1/2), (2, 1/2)]
    ##thr2 = [(2, 1/4), (3, 3/4)]
    thr1 = [(0.1, 1./2), (1.0, 1./2)]
    thr2 = [(3., 1/4), (4.2, 3/4.)]
    thr3 = [(0.5, 1/4.), (1.2, 3./4)]
    thr4 = [(20., 1.)]
    thr = [thr1, thr2, thr3, thr4]

    experSp = Spectrum('', empty=True)
    experSp.set_confs(exper)
    experSp.normalize()
    thrSp = [Spectrum('', empty=True) for _ in range(len(thr))]
    for i in range(len(thr)):
        thrSp[i].set_confs(thr[i])
        thrSp[i].normalize()
    sol2 = dualdeconv2(experSp, thrSp, .2)
    print('sum:', sum(sol2['probs']+sol2['trash']))
    test = estimate_proportions(experSp, thrSp, MTD=.2, MMD=0.21)

    # Very similar masses can introduce errors due to catastrophical cancellations
    chunk_confs = [[1. + 1e-06, 0.6], [1.4, 0.4]]
    query_confs = [(1.00000, 0.6), (1.5, 0.4)]
    badSp = Spectrum('', empty=True)
    badSp.set_confs(chunk_confs)
    badSp.normalize()

    qSp = Spectrum('', empty=True)
    qSp.set_confs(query_confs)
    qSp.normalize()
    dualdeconv2(badSp, [qSp], 0.003, quiet=False)

    # Solution: use mpf library
    from mpmath import mpf, mp
    mp.dps = 25
    chunk_confs = [[mpf(1.) + mpf(1e-06), 0.6], [1.4, 0.4]]
    query_confs = [(1.00000, 0.6), (1.5, 0.4)]
    badSp = Spectrum('', empty=True)
    badSp.set_confs(chunk_confs)
    badSp.normalize()

    qSp = Spectrum('', empty=True)
    qSp.set_confs(query_confs)
    qSp.normalize()
    dualdeconv2(badSp, [qSp], 0.003, quiet=False)

    # Other tests:
##    experSp2 = Spectrum('', empty=True)
##    fr = exper[:-1].copy()
##    experSp2.set_confs(fr)
##    experSp2.normalize()
##    sol22 = dualdeconv2(experSp2, thrSp[:2], .2)
##    print('sum:', sum(sol22['probs']+sol22['trash']))
##
##    global_mass_axis = set(x[0] for x in experSp2.confs)
##    global_mass_axis.update(x[0] for s in thrSp for x in s.confs)
##    global_mass_axis = sorted(global_mass_axis)
##
##    thr_conf_iters = [list(intensity_generator(t.confs, global_mass_axis)) for t in thrSp]
##    thr_conf_iters = [[(m, i) for m, i in zip(global_mass_axis, cnflist) if i>0] for cnflist in thr_conf_iters]
##
##    exper2 = [(0, 1/4), (1.1, 1/6), (2.2, 5/24), (3.1, 1/8), (4, 1/4)]
##    sp2 = Spectrum('', empty=True)
##    sp2.set_confs(exper2)
##    sp2.normalize()
##    noise = Spectrum('', empty=True)
##    noise.set_confs([exper2[2]])
##    noise.normalize()
##    sp2.WSDistance(thrSp[0]*0.35 + thrSp[1]*0.5 + noise*0.125)
##    dualdeconv2(sp2, thrSp[:2], 1.)
##
##    thr1 = [(1., 1.)]
##    thr2 = [(2., 1.)]
##    exper = [(1., 0.4), (1.5, 0.2), (2, 0.4)]
##    thr = [thr1, thr2]
##    experSp = Spectrum('', empty=True)
##    experSp.set_confs(exper)
##    experSp.normalize()
##    thrSp = [Spectrum('', empty=True) for _ in range(len(thr))]
##    for i in range(len(thr)):
##        thrSp[i].set_confs(thr[i])
##        thrSp[i].normalize()
##    sol2 = dualdeconv2(experSp, thrSp, .5)



#! /usr/bin/python3
from masserstein import Spectrum
from getopt import getopt
from functools import reduce
import numpy as np
import sys


doc="""NAME:
    WSDistance.py

USAGE:
    python WSDistance.py [OPTIONS] FILE1 FILE2

EXAMPLES:
    python WSDistance.py examples/ethanol.txt examples/acetic_acid.txt
    python WSDistance.py -t 1. -s examples/ethanol.txt examples/propane.txt

DESCRIPTION:
    Computes the Wasserstein distance between two spectra.
    The two spectra need to be supplied as peak lists in separate files (FILE1, FILE2).
    Each file should be composed of two tab-separated columns.
    The first column corresponds to m/z values, the second one to peak intensity (ion current).
    A header can be present, starting with a hash sign #.
    Any line starting with # will be ommitted.
    The supplied spectra do not need to have normalized peak intensities.
    The normalization will be performed by the program.

RETURNS:
    By default, the Wasserstein distance is printed to stdout, together with the program configuration.
    Optionally, mass transport can be printed to stdout as well. The mass transport is
    represented as a table with three columns, corresponding to the origin m/z, the target m/z,
    and the amount of transported intensity.

OPTIONS:
    -s
        Print the optimal mass transport scheme.
    -t: float
        The total intensity that is to remain in a spectrum after denoising.
        Default: 0.99, which means that peaks corresponding to at most 0.01 of the intensity
        will be removed.
        Setting this parameter to 1 disables denoising.
    -h
        Print this message and exit.

CONTACT:
    If you encounter any problems during use of this application, please email me at m_ciach@student.uw.edu.pl.
"""

def main():

    thr = 0.99
    norm = True
    print_transport = False

    opts, args = getopt(sys.argv[1:], 'hst:')
    if not args:
        print(doc)
        quit()

    for opt, arg in opts:
        if opt == '-t':
            thr = float(arg)
            if not 0 <= thr <= 1:
                raise ValueError("Improper threshold value: %f" % thr)
        if opt == '-h':
            print(doc)
            quit()
        if opt == '-s':
            print_transport = True

    sp1, sp2 = args

    print("Spectrum 1:", sp1)
    print("Spectrum 2:", sp2)
    print("Intensity cutoff:", thr)

    sp1 = open(sp1).readlines()
    sp2 = open(sp2).readlines()

    sp1 = [l.strip() for l in sp1]
    sp2 = [l.strip() for l in sp2]

    sp1 = [list(map(float, l.split())) for l in sp1 if l and l[0] != '#']
    sp2 = [list(map(float, l.split())) for l in sp2 if l and l[0] != '#']

    if norm:
        sum1 = sum(l[1] for l in sp1)
        sp1 = [(l[0], l[1]/sum1) for l in sp1]
        sum2 = sum(l[1] for l in sp2)
        sp2 = [(l[0], l[1]/sum2) for l in sp2]

    if thr < 1:
        order1 = sorted([x for x in enumerate((l[1] for l in sp1))], key=lambda y: y[1])  # ordering of intensities
        cmsm1 = np.cumsum([x[1] for x in order1])  # cumsum of ordered intensities
        to_remove1 = set([o[0] for o, c in zip(order1, cmsm1) if c < 1-thr])  # indices of peaks below threshold
        sp1 = [l for i, l in enumerate(sp1) if i not in to_remove1]  # denoised spectrum

        order2 = sorted([x for x in enumerate((l[1] for l in sp2))], key=lambda y: y[1])
        cmsm2 = np.cumsum(order2)
        to_remove2 = set([o[0] for o, c in zip(order2, cmsm2) if c < 1-thr])
        sp2 = [l for i, l in enumerate(sp2) if i not in to_remove2]

    Spectrum1 = Spectrum("", empty=True)
    Spectrum1.set_confs(sp1)
    Spectrum1.normalize()
    Spectrum2 = Spectrum("", empty=True)
    Spectrum2.set_confs(sp2)
    Spectrum2.normalize()

    W = Spectrum1.WSDistance(Spectrum2)

    print()
    print("Wasserstein distance:")
    print(W)

    if print_transport:
        mvs = list(Spectrum1.WSDistanceMoves(Spectrum2))
        print("Optimal transport scheme:")
        for m in mvs:
            if m[2] > 1e-6:
                print('\t'.join(map(str, m)))

if __name__ == "__main__":
    main()



import random
import numpy as np
import re


def _weighted_var(x, weights):
    return np.average((x - np.average(x, weights=weights)) ** 2, weights=weights)


def goodness_of_fit(transport_plan, metric="L1"):
    # todo: docs
    if not isinstance(transport_plan, np.ndarray):
        transport_plan = np.array(list(transport_plan))

    assert (
        transport_plan[:, 2] >= 0
    ).all(), "Negative values passed in transport plan!"
    transport_plan = transport_plan[transport_plan[:, 2] > 0]
    transport_distance = transport_plan[:, 1] - transport_plan[:, 0]
    transport_mass = transport_plan[:, 2]

    if metric == "L1":
        return np.average(np.abs(transport_distance), weights=transport_mass)
    elif metric == "L2":
        return np.average(transport_distance ** 2, weights=transport_mass) ** 0.5
    elif metric == "std":
        return _weighted_var(transport_distance, transport_mass) ** 0.5
    else:
        raise KeyError("Unknown metric, has to be one of L1, L2, std.")


def generate_random_spectrum(target_mass, elements):
    elements = list(set(elements) - set("H"))
    random.shuffle(elements)
    composition = dict()
    while elements:
        e = elements.pop()
        m = Spectrum(e).confs[0][0]
        if m > target_mass:
            continue
        u = random.randint(0, target_mass // m)
        composition[e] = u
        target_mass -= m * u
    H_mass = Spectrum("H").confs[0][0]
    composition["H"] = np.round(target_mass / H_mass)
    spectrum = Spectrum(
        "".join(f"{k}{int(v)}" for k, v in sorted(composition.items()) if v != 0)
    )
    return spectrum


def get_composition(formula):
    composition = dict()
    for element_and_potential_number in re.findall("[A-Z][a-z]*\d*", formula):
        element = re.findall("[A-Z][a-z]*", element_and_potential_number)[0]
        potential_number = re.findall("\d+$", element_and_potential_number)
        number = int(potential_number[0]) if potential_number else 1
        composition[element] = number
    return composition


def calculate_metrics_for_randomizations(spectrum, elements=None, n_replications=100, metric="L1"):
    if not elements:
        elements = list(get_composition(spectrum.formula).keys())
    mass = spectrum.confs[0][0]
    metrics = []
    for _ in range(n_replications):
        random_spectrum = generate_random_spectrum(mass, elements)
        transport_plan = spectrum.WSDistanceMoves(random_spectrum)
        metric_value = goodness_of_fit(transport_plan, metric=metric)
        metrics.append(metric_value)

    return metrics


# tmp tests
from masserstein.spectrum import Spectrum

s1 = Spectrum("H")
s2 = Spectrum("C")
transport_plan = s1.WSDistanceMoves(s2)
goodness_of_fit(transport_plan)

for i in range(3):
    s = generate_random_spectrum(100, list("HCO"))
    print(s.formula, s.confs[0][0], get_composition(s.formula))

metrics = calculate_metrics_for_randomizations(
    Spectrum("C8H10N4O2", label="Caffeine"), n_replications=10000, metric="std"
)
import matplotlib.pyplot as plt

plt.hist(metrics, bins=20)
plt.savefig("fig.png")



from collections import Counter
import re


aminoacids = {
'A' : 'C3H5N1O1',
'C' : 'C3H5N1O1S1',
'D' : 'C4H5N1O3',
'E' : 'C5H7N1O3',
'F' : 'C9H9N1O1',
'G' : 'C2H3N1O1',
'H' : 'C6H7N3O1',
'I' : 'C6H11N1O1',
'K' : 'C6H12N2O1',
'L' : 'C6H11N1O1',
'M' : 'C5H9N1O1S1',
'N' : 'C4H6N2O2',
'O' : 'C12H21N3O3',
'P' : 'C5H7N1O1',
'Q' : 'C5H8N2O2',
'R' : 'C6H12N4O1',
'S' : 'C3H5N1O2',
'T' : 'C4H7N1O2',
'U' : 'C3H5N1O1Se1',
'V' : 'C5H9N1O1',
'W' : 'C11H10N2O1',
'Y' : 'C9H9N1O2',
' ' : '',
'\n' : ''
}

daminoacids = {}

# From now we are able to model all changes which happen to peptides!
# Notation of changes is a modX proposed by pyteomics.

# Variable modifications of peptides
# N -> D # deamidation
# Q -> E # deamidation
# Q -> pyroglutamic acid # deamidation
# M -> methionine sulfoxide # oxidation
modifications = [("deaN", Counter({"H": -1, "N": 1, "O": 1})),
                 ("deaeQ", Counter({"H": -1, "N": -1, "O": 1})),
                 ("deapQ", Counter({"H": -3, "N": -1})),
                 ("oxM", Counter({"O": 1})),
                 ("carC", Counter({"C": 2, "H": 4, "N": 2, "O": 1}))]

for symbol, formula in aminoacids.items():
    symbols = re.findall("\D+", formula)
    atom_counts = [int(x) for x in re.findall("\d+", formula)]
    C = Counter(dict(zip(symbols, atom_counts)))
    daminoacids[symbol] = C

def get_protein_counter(seq, add_water=True):
    cnts = Counter({k : v for k, v in Counter(seq).items() if v > 0 and k in aminoacids})
    return aacnt_to_elecnt(cnts, add_water=add_water)

def get_protein_formula(seq, add_water=True):
    c = get_protein_counter(seq, add_water)
    # Apply modifications
    for mod_desc, mod_val in modifications:
        mod_cnt = seq.count(mod_desc)
        c += {k : v*mod_cnt for k, v in mod_val.items()}
    return ''.join(sym+str(count) for sym,count in sorted(c.items()))

def aacnt_to_elecnt(cnts, add_water = True):
    if add_water:
        ret = Counter({"H":2, "O":1})
    else:
        ret = Counter()
    for aa, cnt in cnts.items():
        for ele, ecnt in daminoacids[aa].items():
            ret[ele] += ecnt*cnt
    return ret

if __name__ == "__main__":
    import sys
    print(get_protein_formula(sys.argv[1]))



#! /usr/bin/python3
from masserstein import Spectrum
from masserstein import estimate_proportions
from getopt import getopt
import numpy as np
import sys
import re

doc = """NAME:
    WSDeconv

USAGE:
    python3 WSDeconv [OPTIONS] MASS_SPECTRUM MOLECULE_LIST [OUTPUT]

EXAMPLES:
    python3 WSDeconv examples/small_molecule_spectrum.txt "C2H5OH,C3H8,C2H4"
    python3 WSDeconv examples/small_molecule_spectrum.txt examples/small_molecule_list.txt Small_example
    python3 WSDeconv examples/protein_spectrum.txt examples/protein_molecule_list.txt Proteins_example

DESCRIPTION:
    Deconvolves overlapping isotopic distributions and returns their proportions.
    MASS_SPECTRUM file contains the subject spectrum in a peak list format.
    MOLECULE_LIST is either an elemental formula, or a file with a list of
    elemental formulas of ions which proportions in the spectrum are to be estimated.
    Lines starting with a hash # in any file are treated as comments and skipped.
    OUTPUT is the filename prefix of output files. The program writes out the following files:
    OUTPUT_proportions.txt, which stores the list of estimated proportions;
    OUTPUT_denoised.txt, which stores the denoised experimental spectrum;
    OUTPUT_fitted.txt, which stores a fitted theoretical spectrum (a linear combination
    of theoretical spectra of input molecules with optimal proportions);
    OUTPUT_transport.txt, which stores an optimal transport plan from
    the experimental spectrum to the fitted theoretical spectrum;
    OUTPUT_log.txt, which stores the information about the program run.
    If this argument is ommited, the program run information and
    molecule proportions are written to standard output.

    Each elemental formula should consist of three parts:
    a formula of the neutral state molecule, followed by charge sign (plus/minus),
    followed by adduct formula. The charge is determined by the number of adducts atoms.
    If the charge is negative, the adducts are subtracted from the formula.
    An example formula is NH3 + H for ammonium.
    If only the neutral part is supplied, it is assumed that it represents a complete formula
    of a singly-charged ion. Therefore, writing NH3 + H is equivalent to NH4,
    and COOH - H is equivalent to COO. However, simply writing 'NH3 ' will inform the program
    that this is a singly-charged modecule!

OPTIONS:
    -h
        Print this message and exit.
    -p: float in [0, 1], default: 0.95
        Theoretical isotopic envelope coverage.
        This is the percentage of total peak intensity included in the theoretical spectra
        generated by IsoSpec.
        Larger values lead to more accurate theoretical isotopic envelopes.
        This may improve results accuracy, but increase the computational complexity.
        If the experimental spectrum is processed (e.g. peak-picked and/or deisotoped), this
        value needs to be adjusted accordingly to reflect the observed signal.
    -t: float, default: 0.01
        The denoising penalty, interpretable as a maximum feasible distance of ion transport
        or as the cost of assuming that there is no signal in the subject spectrum. 
        The proportion of signal explained by a given theoretical envelope
        is mostly computed from ion current around theoretical peaks within this margin.
        Experimental ion current that cannot be explained by ion current transport
        is treated as background or chemical noise, so that the method performs simultaneous
        proportion estimation and denoising.
        Setting this value to -1 disables denoising.
    -c: float, default: 1e-12
        Minimum detectable ion current. If a theoretical isotopic envelope matches less than
        this proportion of experimental ion current, it is filtered out during preprocessing
        and it's proportion is assumed to be zero.
        Defaults to 0, which effectively disables filtering.
    -d: float, default: 2.1
        Mode Matching Distance. If the highest peak of a theoretical isotopic envelope
        does not match any experimental peak within this distance, this envelope is
        filtered out during preprocessing and it's proportion is assumed to be zero.
        Set to -1 to disable filtering.
    -s
        Suppress writing additional output files - write out only proportions.
    -v
        Print detailed diagnostic messages.

CONTACT:
    If you encounter any problems during use of this application, please email me at m_ciach@student.uw.edu.pl.
"""

to_add_when_implemented = """
    Elemental formulas may be also in the form of 'molecular regular expressions',
    e.g. CH3(CH2)[2,5-7]NH2 + H[0-1]Na[0-1]. This will expand the CH2 part either 2 or from 5 to 7
    times and, for each chain length, will add a hydrogen, a sodium, or both atoms.
    As before, the first part is assumed represent a neutral state of the molecule.
    """

mass_warning = """
WARNING: Detected a large-distance mass transport between masses %.2f and %.2f.
Please report this to the authors.
"""

def main():
    penalty = 0.1  # denoising penalty
    prob = 0.999  # minimum theoretical envelope coverage
    MMD = 2.1  # maximum mode distance
    MDC = 1e-12 # minimum detectable current
    only_proportions = False
    verbose = False

    opts, args = getopt(sys.argv[1:], 'hp:t:c:d:sv')

    if not args:
        print(doc)
        quit()

    for opt, arg in opts:
        if opt == '-h':
            print(doc)
            quit()
        if opt == '-p':
            prob = float(arg)
            assert 0 <= prob <= 1, 'Improper isotopic envelope coverage value: %f' % prob
        if opt == '-t':
            penalty = float(arg)
            assert penalty == -1 or penalty >= 0, 'Improper maximum transport distance: %f' % thr
        if opt == '-c':
            MDC = float(arg)
            assert MDC >= 0, 'Improper Minimum Detectable Current value: %f' % MDC
        if opt == '-d':
            MMD = float(arg)
            assert MMD == -1 or MMD >= 0, 'Improper Maximum Mode Distance: %f' % MMD
        if opt == '-v':
            verbose = True
        if opt == '-s':
            only_proportions = True

    spectrum, molecules = args[:2]
    try:
        output = args[2]
    except IndexError:
        output = None

    LOG = 'WsDeconv initialized.\n'
    LOG += "Experimental spectrum: " + spectrum + '\n'
    LOG += "Molecule list: " + molecules + '\n'
    LOG += "Theoretical envelope coverage: " + str(prob) + '\n'
    if penalty == -1:
        LOG += "Maximum transport distance (denoising penalty): infinite" + '\n'
    else:
        LOG += "Maximum transport distance (denoising penalty): " + str(penalty) + '\n'
    if MMD == -1:
        LOG += 'Mode matching filtering disabled\n'
    else:
        LOG += 'Minimum mode matching distance: ' + str(MMD) + '\n'
    if MDC == 0:
        LOG += 'Minimum current coverage filtering disabled\n'
    else:
        LOG += 'Minimum current coverage: ' + str(MDC) + '\n'

    print(LOG)

    # Parse molecule list & construct list of theoretical spectra:
    def parse_mol_formula(mstr):
        """
        Expands the molecular regular expression mstr to a list of formulas.
        Each element of the returned list is a tuple with a full formula, the adduct formula, and charge.
        """
        formulas = []
        charge_sign = re.findall('[+-]', mstr)
        if len(charge_sign) == 1:
            charge_sign = charge_sign[0]
            neutral, adduct = mstr.split(charge_sign)
            neutral = neutral.strip()
            adduct = adduct.strip()
            adduct_parsed = re.findall('([A-Z][a-z]*)([0-9]*)', adduct)
            adduct, charge = adduct_parsed[0]
            charge = int(charge)
    ##        for e, n in adduct_parsed:
    ##            charge += int(n) if n else 1
        elif len(charge_sign) == 0:
            neutral, adduct = mstr.strip(), None
            charge = 1
        else:
            raise ValueError('Improper charge signs:' + str(charge_sign))
        return (neutral, adduct, charge)

    try:
        molecules = open(molecules).readlines()
    except IOError:
        molecules = molecules.split(',')
    if verbose:
        print('Read molecules:')
        print(molecules)
    molecules = [m.strip() for m in molecules if m and m[0] != '#']
    parsed_molecules = [parse_mol_formula(m) for m in molecules]
    if verbose:
        print('Molecule\tAdduct\tCharge')
        for n,a,c in parsed_molecules:
            print(n,a,c,sep='\t')
    thr_spctrs = [Spectrum(f, threshold = 1-prob, intensity=1.0 , charge=c, adduct=a) for f, a, c in parsed_molecules]
    for s in thr_spctrs:
        s.normalize()

    # Parsing spectrum:
    peaklist = open(spectrum).readlines()
    peaklist = [list(map(float, l.strip().split())) for l in peaklist if l and l[0] != '#']

    # Spectrum initialization:
    spectrum = Spectrum("", empty=True)
    spectrum.set_confs(peaklist)
    del peaklist

    # Normalize and obtain mz range:
    spectrum.normalize()
    mass_range = spectrum.confs[-1][0] - spectrum.confs[0][0]
    if penalty == -1:
        penalty = mass_range + 10.

    # Proportion estimation:
    result = estimate_proportions(spectrum, thr_spctrs, penalty, MDC, MMD, verbose)
    total_signal = sum(result['proportions'])
    result['proportions'] = [p/total_signal for p in result['proportions']]

    # Parsing results:
    if result:
        print('Amount of noise detected: %f' % sum(result['noise']))
        LOG += 'Amount of noise detected: %f' % sum(result['noise']) + '\n'
        # Obtain fitted theoretical spectrum:
        fitted = Spectrum("", empty=True)
        # normalized_proportions = [w/sum(result['proportions']) for w in result['proportions']]
        for s, w in zip(thr_spctrs, result['proportions']):
            fitted += s*w
        fitted.normalize()

        # Obtain denoised spectrum:
        denoised = Spectrum('', empty=True)
        denoised.set_confs([(p[0], p[1]-n ) for p, n in zip(spectrum.confs, result['noise']) if not np.isclose(p[1]-n, 0.)])
        denoised.normalize()

        # Obtain transport plan:
        mvs = [mv for mv in fitted.WSDistanceMoves(denoised) if mv[2] > 1e-06]
        max_mv = max(mvs, key=lambda x: abs(x[1]-x[0]))
    ##    if abs(max_mv[1]-max_mv[0]) > penalty:
    ##        print((mass_warning % (max_mv[0], max_mv[1])))
        wsdist = denoised.WSDistance(fitted)

        print()
        print('Optimal Wasserstein distance with denoising penalty: %f' % (wsdist + penalty*sum(result['noise'])))
        print("Optimal Wasserstein distance: %f" % wsdist)
        print()
        LOG += 'Optimal Wasserstein distance with denoising penalty: %f' % (wsdist + penalty*sum(result['noise'])) + '\n'
        LOG += "Optimal Wasserstein distance: %f" % wsdist + '\n'

        if not output:
            print("Isotopic envelope proportions:")
            for m, d in zip(molecules, result['proportions']):
                print(m + '\t' + str(round(d, 10)))
            print()
        else:
            with open(output+'_proportions.txt', 'w') as h:
                for m, d in zip(molecules, result['proportions']):
                    h.write(m + '\t' + str(d) + '\n')
                print('Optimal molecule proportions written to', output+'_proportions.txt')
            with open(output+'_denoised.txt', 'w') as h:
                for m, i in denoised.confs:
                    h.write(str(m) + '\t'  + str(i) + '\n')
                print('Denoised experimental spectrum written to', output+'_denoised.txt')
            with open(output+'_fitted.txt', 'w') as h:
                for m, i in fitted.confs:
                    h.write(str(m) + '\t' + str(i) + '\n')
                print('Fitted theoretical spectrum written to', output + '_fitted.txt')
            with open(output+'_transport.txt', 'w') as h:
                for em, tm, i in mvs:
                    h.write(str(em) + '\t' + str(tm) + '\t' + str(i) + '\n')
                print('Optimal transport plan written to', output+'_transport.txt')
            with open(output+'_log.txt', 'w') as h:
                h.write(LOG)
                print('Program run information written to', output+'_log.txt')
    else:
        print("Deconvolution failed. Please report this to the authors.")


if __name__ == "__main__":
    main()
